@@ -73,10 +73,11 @@ public final class CreatePnsSubscriptionTable extends AbstractCreateTableImpl {
             "transport VARCHAR(32) CHARACTER SET latin1 NOT NULL," +
             "last_modified BIGINT(64) NOT NULL," +
             "all_flag TINYINT UNSIGNED NOT NULL default '0'," +
+            "expires BIGINT(20) DEFAULT NULL," +
             "PRIMARY KEY (cid, user, token, client)," +
             "UNIQUE KEY `subscription_id` (`id`)" +
             // "INDEX `affiliationIndex` (cid, user, affiliation)" +
-            ") ENGINE=InnoDB DEFAULT CHARSET=utf8 COLLATE=utf8_unicode_ci";
+            ") ENGINE=InnoDB DEFAULT CHARSET=utf8mb4 COLLATE=utf8mb4_unicode_ci";
     }
 
     /**
@@ -90,7 +91,7 @@ public final class CreatePnsSubscriptionTable extends AbstractCreateTableImpl {
             "cid INT4 UNSIGNED NOT NULL," +
             "topic VARCHAR(255) CHARACTER SET latin1 NOT NULL," +
             "PRIMARY KEY (id, topic)" +
-            ") ENGINE=InnoDB DEFAULT CHARSET=utf8 COLLATE=utf8_unicode_ci";
+            ") ENGINE=InnoDB DEFAULT CHARSET=utf8mb4 COLLATE=utf8mb4_unicode_ci";
     }
 
     /**
@@ -104,7 +105,7 @@ public final class CreatePnsSubscriptionTable extends AbstractCreateTableImpl {
             "cid INT4 UNSIGNED NOT NULL," +
             "topic VARCHAR(255) CHARACTER SET latin1 NOT NULL," +
             "PRIMARY KEY (id, topic)" +
-            ") ENGINE=InnoDB DEFAULT CHARSET=utf8 COLLATE=utf8_unicode_ci";
+            ") ENGINE=InnoDB DEFAULT CHARSET=utf8mb4 COLLATE=utf8mb4_unicode_ci";
     }
 
     // ----------------------------------------------------------------------------------------------------------------
