@@ -54,10 +54,12 @@ import java.io.IOException;
 import java.util.ArrayList;
 import java.util.Collection;
 import java.util.HashMap;
+import java.util.Iterator;
 import java.util.LinkedHashMap;
 import java.util.LinkedList;
 import java.util.List;
 import java.util.Map;
+import java.util.Map.Entry;
 import java.util.concurrent.TimeUnit;
 import org.osgi.framework.BundleContext;
 import org.osgi.framework.ServiceReference;
@@ -68,12 +70,12 @@ import com.google.android.gcm.Constants;
 import com.google.android.gcm.Endpoint;
 import com.google.android.gcm.Message;
 import com.google.android.gcm.Message.Priority;
-import com.google.common.cache.Cache;
-import com.google.common.cache.CacheBuilder;
 import com.google.android.gcm.MulticastResult;
 import com.google.android.gcm.Notification;
 import com.google.android.gcm.Result;
 import com.google.android.gcm.Sender;
+import com.google.common.cache.Cache;
+import com.google.common.cache.CacheBuilder;
 import com.openexchange.config.cascade.ComposedConfigProperty;
 import com.openexchange.config.cascade.ConfigView;
 import com.openexchange.config.cascade.ConfigViewFactory;
@@ -237,6 +239,15 @@ public class GcmPushNotificationTransport extends ServiceTracker<GcmOptionsProvi
         return false;
     }
 
+    @Override
+    public void transport(Map<PushNotification, List<PushMatch>> notifications) throws OXException {
+        if (null != notifications && 0 < notifications.size()) {
+            for (Entry<PushNotification, List<PushMatch>> entry : notifications.entrySet()) {
+                transport(entry.getKey(), entry.getValue());
+            }
+        }
+    }
+
     @Override
     public void transport(PushNotification notification, Collection<PushMatch> matches) throws OXException {
         if (null != notification && null != matches) {
@@ -272,7 +283,8 @@ public class GcmPushNotificationTransport extends ServiceTracker<GcmOptionsProvi
                 if (!registrationIDs.isEmpty()) {
                     MulticastResult result = null;
                     try {
-                        result = sender.sendNoRetry(getMessage(client, notification), registrationIDs, Endpoint.FCM);
+                        // result = sender.sendNoRetry(getMessage(client, notification), registrationIDs, Endpoint.FCM);
+                        result = sender.send(getMessage(client, notification), registrationIDs, 3, Endpoint.FCM);
 
                         // Log it
                         Object ostr = new Object() {
@@ -280,20 +292,23 @@ public class GcmPushNotificationTransport extends ServiceTracker<GcmOptionsProvi
                             @Override
                             public String toString() {
                                 StringBuilder sb = new StringBuilder(registrationIDs.size() * 16);
-                                for (String registrationID : registrationIDs) {
-                                    sb.append(registrationID).append(", ");
+                                Iterator<String> it = registrationIDs.iterator();
+                                sb.append(it.next());
+                                while (it.hasNext()) {
+                                    sb.append(", ").append(it.next());
                                 }
-                                sb.setLength(sb.length() - 2);
                                 return sb.toString();
                             }
                         };
-                        LOG.info("Sent notification \"{}\" via transport '{}' for user {} in context {} to registration ID(s): {}", notification.getTopic(), ID, I(notification.getUserId()), I(notification.getContextId()), ostr);
+                        if (null != result) {
+                            LOG.info("Sent notification \"{}\" via transport '{}' for user {} in context {} to registration ID(s): {}", notification.getTopic(), ID, I(notification.getUserId()), I(notification.getContextId()), ostr);
+                            LOG.debug("{}", result);
+                        } else {
+                            LOG.info("Failed to send notification \"{}\" via transport '{}' for user {} in context {} to registration ID(s): {}", notification.getTopic(), ID, I(notification.getUserId()), I(notification.getContextId()), ostr);
+                        }
                     } catch (IOException e) {
                         LOG.warn("Error publishing push notification", e);
                     }
-                    if (null != result) {
-                        LOG.debug("{}", result);
-                    }
                     /*
                      * process results
                      */
